using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using Utils;
using SplineUtils;

namespace SplineUtils
{
	public enum eOrientationMode { NODE = 0, TANGENT, NONE }

	[AddComponentMenu("Splines/Spline Controller")]
	[RequireComponent(typeof(SplineInterpolator))]

	public class SplineController : MonoBehaviour
	{
		[HideInInspector]
		public float Duration = 10;
		public eOrientationMode OrientationMode = eOrientationMode.NODE;
		public eWrapMode WrapMode = eWrapMode.ONCE;
		[HideInInspector]
		public bool AutoStart = true;
		[HideInInspector]
		public bool AutoClose = true;
		[HideInInspector]
		public AnimationCurve SpeedControlClip = null;
		[HideInInspector]
		public GameObject AnimationTestObject = null;
		[HideInInspector]
		public bool ShowResampledSpline = false;
		
		[HideInInspector]
		public Color color;

		[HideInInspector]
		public List<CoordStruct> mTransforms;
		[HideInInspector]
		public bool dirty = true;
		[HideInInspector]
		public List<SplineNode> mSpline;

		private SplineInterpolator mSplineInterp;
		private List<SplineNode> mResampledSpline;
		private List<SplineInterpolatorMK2.TableData> mLookupTable;
		private Vector3 mPrevPos;

		void OnDrawGizmosSelected()
		{
			List<CoordStruct> trans = mTransforms;

			if ( trans.Count < 2 )
				return;

			SplineInterpolator interp = GetComponent( typeof(SplineInterpolator) ) as SplineInterpolator;
			SetupSplineInterpolator( interp, trans );
			interp.StartInterpolation( null, false, WrapMode );

			if ( mSpline == null )
				mSpline = SplineInterpolatorMK2.Setup( mTransforms, OrientationMode, WrapMode, AutoClose );

			// Draw gizmo line
			List<SplineNode> mNodes = resampledSpline;

			Vector3 prevPos = trans[0].position;
			for ( int c = 1; c <= 100; c++ )
			{
				float currTime 	= (float)c / 100.0f;
				Vector3 currPos = interp.GetHermiteAtTime( currTime );
				currPos.y 		+= 0.1f;
				Gizmos.color 	= color;
				Gizmos.DrawLine( prevPos, currPos );
				prevPos = currPos;
			}

			if ( ShowResampledSpline )
			{
				for ( int i = 1; i < mNodes.Count - 1; i += 7 )
				{
					Gizmos.color = Color.yellow;
					Gizmos.DrawSphere( mNodes[i].Point, 0.075f );
				}
			}
		}


		public void ForceRecalculate()
		{
			dirty = true;
			mResampledSpline = null;
		}

		public List<SplineNode> resampledSpline
		{
			get 
			{ 
				if ( mResampledSpline == null || dirty )
				{
					dirty = false;
					mSpline = SplineInterpolatorMK2.Setup( mTransforms, OrientationMode, WrapMode, AutoClose );
					mResampledSpline = SplineInterpolatorMK2.ResampleCurve( mSpline, 0.05f, OrientationMode );
					mResampledSpline = SplineInterpolatorMK2.Setup( mResampledSpline, OrientationMode, WrapMode, AutoClose );
				}

				return mResampledSpline;
			}
		}
	

		void Start()
		{
			mSplineInterp = GetComponent( typeof( SplineInterpolator ) ) as SplineInterpolator;
		}


		void SetupSplineInterpolator( SplineInterpolator interp, List<CoordStruct> trans )
		{
			interp.Reset();

			float step = (AutoClose) ? Duration / trans.Count :
									   Duration / ( trans.Count - 1 );

			int c;
			for ( c = 0; c < trans.Count; c++ )
			{
				if ( OrientationMode == eOrientationMode.NODE )
				{
					interp.AddPoint( trans[c].position, Quaternion.Euler( trans[c].rotation ), step * c, new Vector2( 0, 1 ) );
				}
				else if ( OrientationMode == eOrientationMode.TANGENT )
				{
					Quaternion rot;
					if ( c != trans.Count - 1 )
						rot = Quaternion.LookRotation( trans[c + 1].position - trans[c].position, Vector3.up ); //trans[c].up );
					else if ( AutoClose )
						rot = Quaternion.LookRotation( trans[0].position - trans[c].position, Vector3.up ); //trans[c].up );
					else
						rot = Quaternion.Euler( trans[c].rotation );

					interp.AddPoint( trans[c].position, rot, step * c, new Vector2( 0, 1 ) );
				}
			}

			if ( AutoClose )
				interp.SetAutoCloseMode( step * c );
		}

		/// <summary>
		/// Starts the interpolation
		/// </summary>
		void FollowSpline()
		{
			if ( mTransforms.Count > 0 )
			{
				SetupSplineInterpolator( mSplineInterp, mTransforms );
				mSplineInterp.StartInterpolation( null, true, WrapMode );
			}
		}
	}
}