using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using Utils;

namespace SplineUtils
{
	public class SplineInterpolatorMK2 : MonoBehaviour 
	{
		public static string mState = "";

		public struct TableData
		{
			public float 	t;
			public float 	arcLength;
			public int 		idx;

			public TableData( float time, float a, int index ) { t = time; arcLength = a; idx = index; }
			public TableData( float time, float a ) { t = time; arcLength = a; idx = 0; }
		}

		public struct SplineEvalData
		{
			public Vector3 		pos;
			public Quaternion 	rot;

			public SplineEvalData( Vector3 position, Quaternion rotation ) { pos = position; rot = rotation; }
		}

		public static List<SplineNode> Setup( List<SplineNode> nodes, eOrientationMode orientationMode, eWrapMode wrapMode, bool AutoClose )
		{
			List<CoordStruct> trans = new List<CoordStruct>(nodes.Count);

			for ( int i = 0; i < nodes.Count; i++ )
			{
				trans.Add( new CoordStruct( nodes[i].Point, nodes[i].Rot.eulerAngles ) );
			}

			return Setup( trans, orientationMode, wrapMode, AutoClose );
		}


		public static List<SplineNode> Setup( List<CoordStruct> trans, eOrientationMode orientationMode, eWrapMode wrapMode, bool AutoClose )
		{
			int nodeCount = trans.Count;
			if ( AutoClose )
			{
				nodeCount += 1;
			}

			List<SplineNode> mNodes = new List<SplineNode>(nodeCount);
			eEndPointsMode mEndPointsMode = eEndPointsMode.AUTO;
			
			bool mRotations = true;

			mState = wrapMode == eWrapMode.ONCE ? "Once" : "Loop";

			//////////////////////////////////////////////////////////////////
			
			float step = ( AutoClose ) ? 1.0f / trans.Count : 1.0f / ( trans.Count - 1 );

			int c;
			for ( c = 0; c < trans.Count; c++ )
			{
				if ( orientationMode == eOrientationMode.NODE )
				{
					mNodes.Add( new SplineNode( trans[c].position, Quaternion.Euler( trans[c].rotation ), step * c, new Vector2( 0, 1 ) ) );
				}
				else if ( orientationMode == eOrientationMode.TANGENT )
				{
					Quaternion rot;
					if (c != trans.Count - 1)
						rot = Quaternion.LookRotation( trans[c + 1].position - trans[c].position, Vector3.up );
					else if ( AutoClose )
						rot = Quaternion.LookRotation( trans[0].position     - trans[c].position, Vector3.up );
					else
						rot = Quaternion.Euler( trans[c].rotation );
					
					mNodes.Add( new SplineNode( trans[c].position, rot, step * c, new Vector2( 0, 1 ) ) );
				}
			}
			
			if ( AutoClose )
			{
				float joiningPointTime = step * c;
				mEndPointsMode = eEndPointsMode.AUTOCLOSED;
				
				mNodes.Add(new SplineNode(mNodes[0] as SplineNode));
				mNodes[mNodes.Count - 1].Time = joiningPointTime;
				
				Vector3 vInitDir = (mNodes[1].Point - mNodes[0].Point).normalized;
				Vector3 vEndDir = (mNodes[mNodes.Count - 2].Point - mNodes[mNodes.Count - 1].Point).normalized;
				float firstLength = (mNodes[1].Point - mNodes[0].Point).magnitude;
				float lastLength = (mNodes[mNodes.Count - 2].Point - mNodes[mNodes.Count - 1].Point).magnitude;
				
				SplineNode firstNode = new SplineNode(mNodes[0] as SplineNode);
				firstNode.Point = mNodes[0].Point + vEndDir * firstLength;
				
				SplineNode lastNode = new SplineNode(mNodes[mNodes.Count - 1] as SplineNode);
				lastNode.Point = mNodes[0].Point + vInitDir * lastLength;
				
				mNodes.Insert(0, firstNode);
				mNodes.Add(lastNode);
			}

			/////////////////////////////////////////////
			if ( mNodes.Count < 2 )
				throw new System.Exception("Invalid number of points");
			
			if ( mRotations )
			{
				for ( c = 1; c < mNodes.Count; c++ )
				{
					SplineNode node = mNodes[c];
					SplineNode prevNode = mNodes[c - 1];
					
					// Always interpolate using the shortest path -> Selective negation
					if ( Quaternion.Dot( node.Rot, prevNode.Rot ) < 0 )
					{
						node.Rot.x = -node.Rot.x;
						node.Rot.y = -node.Rot.y;
						node.Rot.z = -node.Rot.z;
						node.Rot.w = -node.Rot.w;
					}
				}
			}
			
			if ( mEndPointsMode == eEndPointsMode.AUTO )
			{
				mNodes.Insert( 0, mNodes[0] );
				mNodes.Add( mNodes[mNodes.Count - 1] );
			}
			else if ( mEndPointsMode == eEndPointsMode.EXPLICIT && ( mNodes.Count < 4 ) )
				throw new System.Exception("Invalid number of points");

			return mNodes;
		}
	

		public static SplineEvalData EvalAtTime( List<SplineNode> mNodes, float time )
		{
			int mCurrentIdx = 0;

			// We advance to next point in the path
			while ( time >= mNodes[mCurrentIdx + 1].Time )
			{
				if ( mCurrentIdx < mNodes.Count - 3 )
					mCurrentIdx++;
				else
					return new SplineEvalData( mNodes[mNodes.Count -2].Point, mNodes[mNodes.Count - 2].Rot );
			}

			// Calculates the t param between 0 and 1
			float param = ( time - mNodes[mCurrentIdx].Time ) / ( mNodes[mCurrentIdx + 1].Time - mNodes[mCurrentIdx].Time );
			return new SplineEvalData( GetHermiteInternal( mCurrentIdx, param, mNodes ), GetSquad( mCurrentIdx, param, mNodes ) );
		}


		static Vector3 GetHermiteInternal( int idxFirstPoint, float t, List<SplineNode> mNodes )
		{
			float t2 = t * t;
			float t3 = t2 * t;

			Vector3 P0 = mNodes[ idxFirstPoint - 1 ].Point;
			Vector3 P1 = mNodes[ idxFirstPoint     ].Point;
			Vector3 P2 = mNodes[ idxFirstPoint + 1 ].Point;
			Vector3 P3 = mNodes[ idxFirstPoint + 2 ].Point;

			float tension = 0.5f;	// 0.5 = equivalant to catmull-rom 

			Vector3 T1 = tension * ( P2 - P0 );
			Vector3 T2 = tension * ( P3 - P1 );

			float Blend1 = 2 * t3 - 3 * t2 + 1;
			float Blend2 = -2 * t3 + 3 * t2;
			float Blend3 = t3 - 2 * t2 + t;
			float Blend4 = t3 - t2;

			return Blend1 * P1 + Blend2 * P2 + Blend3 * T1 + Blend4 * T2;
		}


		static Quaternion GetSquad( int idxFirstPoint, float t, List<SplineNode> mNodes )
		{
			Quaternion Q0 = mNodes[ idxFirstPoint - 1 ].Rot;
			Quaternion Q1 = mNodes[ idxFirstPoint     ].Rot;
			Quaternion Q2 = mNodes[ idxFirstPoint + 1 ].Rot;
			Quaternion Q3 = mNodes[ idxFirstPoint + 2 ].Rot;

			Quaternion T1 = MathUtils.GetSquadIntermediate( Q0, Q1, Q2 );
			Quaternion T2 = MathUtils.GetSquadIntermediate( Q1, Q2, Q3 );

			return MathUtils.GetQuatSquad( t, Q1, Q2, T1, T2 );
		}
		

		static float ArcLength( List<SplineNode> mNodes, float alpha, float beta )
		{
			float gamma 	= ( alpha + beta ) * 0.5f;
			Vector3 a 		= EvalAtTime( mNodes, alpha ).pos;
			Vector3 b 		= EvalAtTime( mNodes, beta  ).pos;
			Vector3 c 		= EvalAtTime( mNodes, gamma ).pos;
			Vector3 cPrime	= ( a + b ) * 0.5f;

			// If we're within a margin of error, accept result. Otherwise, half the distance and try again. 
			if ( (c - cPrime).sqrMagnitude / ( a - b ).sqrMagnitude > 0.025f )
				return ArcLength( mNodes, alpha, gamma ) + ArcLength( mNodes, gamma, beta );
			else
				return Vector3.Distance( a, b );
		}
		

		static float ArcLength( List<SplineNode> mNodes, float t )
		{
			return ArcLength( mNodes, 0.0f, t );
		}


		public static List<SplineNode> ResampleCurve( List<SplineNode> curve, float spacing, eOrientationMode mode )
		{
			float curveArcLen = ArcLength(curve, 0.0f, 1.0f);
			List<SplineNode> resampledCurve = new List<SplineNode>(Mathf.CeilToInt(curveArcLen/spacing) + 1);

			// Add starting point
			resampledCurve.Add( curve[0] );

			// Make the lookup table
			//int c = 1000;
			// Let's calculate this based on the number of segments instead
			// We don't need this to be very high. 16 is probably plenty high, but 32 is still fast
			int c = curve.Count * 32;
			List<TableData> arcLengthMapping = new List<TableData>(c);
			while ( arcLengthMapping.Count <= c )
			{
				float t = ( arcLengthMapping.Count / (float)c );
				arcLengthMapping.Add( new TableData( t, ArcLength( curve, t ) ) );
			}

			// Resample with even spacing
			for (float f = 0.0f; f < curveArcLen; f += spacing)
			{
				// Get time at given arc-length
				float t = GetInterpolatedTime( f, arcLengthMapping );

				// Weird stuff happens when t ~= 0.5!
				//t = 0.5f;

				// Evaluate curve at interpolated time
				SplineEvalData evalData = EvalAtTime( curve, t );
				resampledCurve.Add( new SplineNode( evalData.pos, evalData.rot, t, Vector2.zero ) );
			}

			return resampledCurve;
		}


		public static float ResampleAtArcLength( float t, List<SplineNode>curve, List<TableData>td )
		{
			return GetInterpolatedTime( t * ArcLength( curve, t ), td );
		}


		static float GetInterpolatedTime( float idealArcLength, List<TableData>td )
		{
			for ( int i = 0; i < td.Count - 1; i++ )
			{
				if ( td[i].arcLength <= idealArcLength && td[i + 1].arcLength >= idealArcLength )
				{
					float diffA = idealArcLength - td[i].arcLength;
					float diffN = td[i + 1].arcLength - td[i].arcLength;

					float result = td[i].t + ( td[i + 1].t - td[i].t ) * diffA / diffN;
					return result;
				}
			}

			return 0; // Not in range
		}

		public static SplineEvalData SampleInterpolatedSpline( float t, List<SplineNode>curve, bool rotateFollowNodes )
		{
			float idealIndex = t * (float)curve.Count;

			for ( int i = 0; i < curve.Count - 1; i++ )
			{
				if ( i <= idealIndex && i + 1 >= idealIndex )
				{
					// TODO: Simplify this
					float diffA = idealIndex - (float)i;
					float diffB = (float)( i + 1 ) - (float)i;
					float t2 = diffA / diffB;

					Vector3 pos  	= Vector3.Lerp(     curve[i].Point,		curve[i + 1].Point, t2 );
					Quaternion rot 	= Quaternion.Slerp( curve[i].Rot, 		curve[i + 1].Rot, 	t2 );

					return new SplineEvalData( pos, rot );
				}
			}

			// Not in range
			return new SplineEvalData( curve[curve.Count - 1].Point, curve[curve.Count - 1].Rot );
		}
	}
}